#!/bin/python
# -*- coding: utf-8 -*-

import sys
import time
import numpy as np
import scipy.optimize as so
from numba import njit, prange


def solve_current(model, shock, XLag, XLastGuess, XPrime, tol):

    func = model["func"]
    pars = np.array(list(model["parameters"].values()))
    stst = np.array(list(model["stst"].values()))

    func_current = lambda x: func(XLag, x, XPrime, stst, shock, pars)

    res = so.root(func_current, XLastGuess, options=model["root_options"])
    err = np.max(np.abs(func_current(res["x"])))

    return res["x"], not res["success"], err > tol


def find_path_linear(model, shock, T, x, use_linear_guess):

    if model.get("lin_pol") is not None:

        stst = np.array(list(model["stst"].values()))
        sel = np.isclose(stst, 0)

        shocks = model.get("shocks") or ()
        tshock = np.zeros(len(shocks))
        if shock is not None:
            tshock[shocks.index(shock[0])] = shock[1]

        x_lin = np.empty_like(x)
        x_lin[0][~sel] = (x[0] / stst - 1)[~sel]
        x_lin[0][sel] = 0

        for t in range(T):
            x_lin[t + 1] = model["lin_pol"][0] @ x_lin[t]

            if not t:
                x_lin[t + 1] += model["lin_pol"][1] @ tshock

        x_lin[:, ~sel] = ((1 + x_lin) * stst)[:, ~sel]

        if use_linear_guess:
            return x_lin.copy(), x_lin
        else:
            return x, x_lin
    else:
        return x, None


def find_pizza(
    model,
    x0=None,
    shock=None,
    T=30,
    init_path=None,
    max_horizon=200,
    max_loops=100,
    max_iter=None,
    tol=1e-5,
    use_linear_guess=False,
    root_options={},
    raise_error=False,
    verbose=True,
):
    """Find the expected trajectory given an initial state. A good strategy is to first set `tol` to a low value (e.g. 1e-3) and check for a good max_horizon. Then, set max_horizon to a reasonable value and let max_loops be high.

    Parameters
    ----------
    model : dict
        model dict as defined/parsed above
    x0 : array
        initial state
    shock : tuple
        shock in period 0 as in `(shock_name_as_str, shock_size)`
    T : int, optional
        number of periods to simulate
    init_path : array, optional
        a first guess on the trajectory. Normally not necessary
    max_horizon : int, optional
        number of periods until the system is assumed to be back in the steady state. A good idea to set this corresponding to the respective problem. Note that a horizon too far away may cause the accumulation of numerical errors.
    max_loops : int, optional
        number of repetitions to iterate over the whole trajectory. Should eventually be high.
    max_iterations : int, optional
        number of iterations. Default is `max_horizon`. It should not be lower than that (and will raise an error). Normally it should not be higher, better use `max_loops` instead.
    tol : float, optional
        convergence criterion
    root_options : dict, optional
        dictionary with solver-specific options to be passed on to `scipy.optimize.root`
    verbose : bool, optional
        degree of verbosity. 0/`False` is silent

    Returns
    -------
    array
        array of the trajectory
    flag
        integer of error flag
    """

    st = time.time()

    if max_iter is None:
        max_iter = max_horizon
    elif max_iter < max_horizon:
        Exception(
            "max_iter should be higher or equal max_horizon, but is %s and %s."
            % (max_iter, max_horizon)
        )

    stst = np.array(list(model["stst"].values()))
    nvars = len(model["variables"])
    shocks = model.get("shocks") or ()

    if root_options:
        model["root_options"] = root_options

    # precision of root finding should be some magnitudes higher than of solver
    if "xtol" not in model["root_options"]:
        model["root_options"]["xtol"] = min(tol / max_horizon, 1e-8)

    x_fin = np.empty((T + 1, nvars))
    x_fin[0] = list(x0) if x0 is not None else stst

    x = np.ones((T + max_horizon + 1, nvars)) * stst
    x[0] = x_fin[0]

    x, x_lin = find_path_linear(model, shock, T + max_horizon, x, use_linear_guess)
    x_lin = x_lin[: T + 1] if x_lin is not None else None

    if init_path is not None:
        x[1 : len(init_path)] = init_path[1:]

    tshock = np.zeros(len(shocks))

    fin_flag = np.zeros(5, dtype=bool)
    old_clock = time.time()

    msgs = (
        ", root finding did not converge",
        ", ftol not reached in root finding",
        ", contains NaNs",
        ", contains infs",
        ", max_iter reached",
    )

    try:
        for i in range(T):

            loop = 1
            cnt = 2
            flag = np.zeros_like(fin_flag)

            while True:

                x_old = x[1].copy()
                imax = min(cnt, max_horizon)

                flag_loc = np.zeros(2, dtype=bool)

                for t in range(imax):

                    if not t and not i and shock is not None:
                        tshock[shocks.index(shock[0])] = shock[1]
                    else:
                        tshock[:] = 0

                    x[t + 1], flag_root, flag_ftol = solve_current(
                        model, tshock, x[t], x[t + 1], x[t + 2], tol
                    )

                    flag_loc[0] |= flag_root
                    flag_loc[1] |= flag_ftol and not flag_root

                flag[2] |= np.any(np.isnan(x))
                flag[3] |= np.any(np.isinf(x))

                if cnt == max_iter:
                    if loop < max_loops:
                        loop += 1
                        cnt = 2
                    else:
                        flag[4] |= True

                err = np.abs(x_old - x[1]).max()

                clock = time.time()
                if verbose and clock - old_clock > 0.5:
                    old_clock = clock
                    print(
                        "   Period{:>4d} | loop{:>5d} | iter.{:>5d} | flag{:>3d} | error: {:>1.8e}".format(
                            i, loop, cnt, 2 ** np.arange(5) @ fin_flag, err
                        )
                    )

                if (err < tol and cnt > 2) or flag.any():
                    flag[:2] |= flag_loc
                    if raise_error and flag.any():
                        mess = [i * bool(j) for i, j in zip(msgs, flag)]
                        raise Exception("Aborting%s" % "".join(mess))
                    fin_flag |= flag
                    break

                cnt += 1

            x_fin[i + 1] = x[1].copy()
            x = x[1:].copy()

    except Exception as error:
        raise type(error)(
            str(error)
            + " (raised in period %s during loop %s for forecast %s steps ahead)"
            % (i, loop, t)
        ).with_traceback(sys.exc_info()[2])

    fin_flag[1] &= not fin_flag[0]
    mess = [i * bool(j) for i, j in zip(msgs, fin_flag)]
    fin_flag = 2 ** np.arange(5) @ fin_flag

    if verbose:
        duration = np.round(time.time() - st, 3)
        print("(find_path:) Pizza done after %s seconds%s." % (duration, "".join(mess)))

    return x_fin, x_lin, fin_flag



#!/bin/python
# -*- coding: utf-8 -*-

import numpy as np
import scipy.optimize as so
from scipy.linalg import block_diag
from .shooting import solve_current


def solve_stst(model, raise_error=True, method=None, tol=1e-8, verbose=True):

    evars = model["variables"]
    func = model["func"]
    par = np.array(list(model["parameters"].values()))
    shocks = model.get("shocks") or ()
    stst = model["stst"]
    stst_np = np.array(list(stst.values()))
    method = method or model["steady_state"].get("method")

    if method not in ("reduction", "aux_function", None):
        raise NotImplementedError(
            "Steady state method must either be 'aux_function' or 'reduction' or `None`."
        )

    if method in ("aux_function", None):

        # create a sequence and ensure that its columns are linearly independent
        shifter_rand = np.arange(len(evars) * len(stst)).reshape(len(evars), len(stst))
        svd_u, _, svd_v = np.linalg.svd(shifter_rand, full_matrices=False)
        shifter = svd_u @ svd_v
        x_ind = np.array([evars.index(v) for v in stst.keys()])

        stst_no_zero = stst_np.copy()
        stst_no_zero[np.isclose(stst_np, 0)] = 1

        def func_stst(x):

            # use the sequence to force root finding to set fixed st.st values
            corr = (x[x_ind] - stst_np) / stst_no_zero

            return func(x, x, x, x, np.zeros(len(shocks)), par, True) + shifter @ corr

        # find stst
        res = so.root(func_stst, model["init"])
        err = np.abs(func_stst(res["x"])).max()
        mess = " ".join(res["message"].replace("\n", " ").split())

    if method == "reduction":

        # create a sequence and ensure that its columns are linearly independent
        shifter_rand = np.ones((len(evars) - len(stst), len(evars)))
        svd_u, _, svd_v = np.linalg.svd(shifter_rand, full_matrices=False)
        shifter = svd_u @ svd_v

        x_ind = np.array([stst.get(v) is not None for v in evars])
        x = np.empty(len(evars))

        def func_stst(x_guess):

            x[x_ind] = stst_np
            x[~x_ind] = x_guess

            return shifter @ func(x, x, x, x, np.zeros(len(shocks)), par, True)

        # find stst
        res = so.root(func_stst, model["init"][~x_ind])
        err = np.abs(func_stst(res["x"])).max()
        mess = " ".join(res["message"].replace("\n", " ").split())

    if err > tol:
        if raise_error and not res["success"]:
            print(res)
            raise Exception(
                "Steady state not found (error is %1.2e). %s See the root finding result above."
                % (err, mess)
            )
        else:
            print(
                "(solve_stst:) Steady state not found (error is %1.2e). %s"
                % (err, mess)
            )
    elif verbose:
        print("(solve_stst:) Steady state found.")

    rdict = dict(zip(evars, res["x"]))
    model["stst"] = rdict
    model["init"] = np.array(list(rdict.values()))
    model["stst_vals"] = np.array(list(rdict.values()))

    return rdict


def solve_linear(
    model,
    x=None,
    eps=1e-5,
    tol=1e-8,
    raise_error=True,
    check_contraction=False,
    lti_max_iter=1000,
    verbose=True,
):
    """Does half-way SGU and uses Klein's method to check for determinancy and solve the system"""

    evars = model["variables"]
    func = model["func"]
    par = np.array(list(model["parameters"].values()))
    shocks = model.get("shocks") or ()
    stst = list(model["stst"].values())
    nshc = len(shocks)

    if x is None:
        x = np.array(stst)

    AA = np.empty((len(stst), len(stst)))
    BB = AA.copy()
    CC = AA.copy()
    DD = np.empty((len(stst), nshc))

    zshock = np.zeros(len(shocks))
    fx = func(x, x, x, x, np.zeros(nshc), par)

    for i in range(len(evars)):

        xerr = x.copy()
        xerr[i] -= eps

        AA[:, i] = (func(x, x, xerr, x, zshock, par) - fx) * x[i] / eps
        BB[:, i] = (func(x, xerr, x, x, zshock, par) - fx) * x[i] / eps
        CC[:, i] = (func(xerr, x, x, x, zshock, par) - fx) * x[i] / eps

    for i in range(len(shocks)):
        cshock = zshock.copy()
        cshock[i] += eps

        DD[:, i] = (func(x, x, x, x, cshock, par) - fx) / eps

    A = np.pad(AA, ((0, nshc), (0, nshc)))
    B = block_diag(BB, np.eye(nshc))
    C = np.block([[CC, DD], [np.zeros((nshc, A.shape[1]))]])

    model["ABC"] = A, B, C

    I = np.eye(len(stst) + nshc)
    Z = np.zeros_like(I)
    P = np.block([[B, A], [I, Z]])
    M = np.block([[C, Z], [Z, -I]])

    mess = ""
    success = True

    try:
        from grgrlib import klein

        _, lam = klein(P, M, len(stst) + nshc, verbose=verbose - 1)
        model["lin_pol"] = -lam[:-nshc, :-nshc], -lam[:-nshc, -nshc:]
        mess = "All eigenvalues are good"

    except ImportError:
        mess = "'grgrlib' not found, could not check eigenvalues"

    except Exception as error:
        success = False
        if raise_error:
            raise error
        else:
            mess = str(error).strip()
            if mess[-1] == ".":
                mess = mess[:-1]

    if check_contraction:
        A = np.linalg.inv(BB) @ AA
        B = np.linalg.inv(BB) @ CC

        Aev = np.abs(np.linalg.eig(A)[0])
        Aev_err = Aev > 1
        Bev = np.abs(np.linalg.eig(B)[0])
        Bev_err = Bev > 1
        flag += Aev_err.any() or Bev_err.any()
        mess += ", but " if success else ""
        if Aev_err.any():
            mess += "%s forward looking EV%s larger than unity (%s)" % (
                Aev_err.sum(),
                "s are" if Aev_err.sum() > 1 else " is",
                *Aev[Aev_err],
            )
        if Aev_err.any() and Bev_err.any():
            mess += " and "
        if Bev_err.any():
            mess += "%s backward looking EV%s larger than unity (%s)" % (
                Bev_err.sum(),
                "s are" if Bev_err.sum() > 1 else " is",
                *Bev[Bev_err],
            )

    if mess and verbose:
        print("(solve_linear:) " + mess + ("" if mess[-1] in ".?!" else "."))

    return success



#!/bin/python
# -*- coding: utf-8 -*-

import sys
import time
import numpy as np
import scipy.optimize as so
from numba import njit, prange
from .shooting import find_path_linear


def stacked_func_plain(x, x0, endpoint, func, horizon, nvars, stst, tshock, zshock, pars):

    out = np.empty((horizon-1)*nvars)
    X = x.reshape((horizon - 1, nvars))

    out[:nvars] = func(x0, X[0], X[1], stst, tshock, pars)
    out[-nvars:] = func(X[-2], X[-1], endpoint, stst, zshock, pars)

    for t in prange(1, horizon - 2):
        out[t*nvars:(t+1)*nvars] = func(X[t - 1], X[t], X[t + 1], stst, zshock, pars)

    return out.flatten()

stacked_func_njit = njit(stacked_func_plain)
stacked_func_njit_parallel = njit(stacked_func_plain, parallel=True)

def find_stack(
    model,
    x0=None,
    shock=None,
    init_path=None,
    horizon=50,
    tol=None,
    use_numba=False,
    use_linear_guess=True,
    use_linear_endpoint=None,
    root_options={},
    verbose=True,
):

    st = time.time()

    stst = np.array(list(model["stst"].values()))
    nvars = len(model["variables"])
    func = model["func"]
    pars = np.array(list(model["parameters"].values()))
    shocks = model.get("shocks") or ()

    if root_options:
        model["root_options"] = root_options

    if "xtol" not in model["root_options"]:
        if tol is None:
            tol = 1e-5
        elif "xtol" in root_options:
            print(
                "(find_path:) Specification of xtol in `root_options` overwrites `tol`"
            )
        model["root_options"]["xtol"] = tol

    x0 = np.array(list(x0)) if x0 is not None else stst
    x = np.ones((horizon + 1, nvars)) * stst
    x[0] = x0

    x_init, x_lin = find_path_linear(model, shock, horizon, x, use_linear_guess)

    if use_linear_endpoint is None:
        use_linear_endpoint = False if x_lin is None else True
    elif use_linear_endpoint and x_lin is None:
        print("(find_path_stacked:) Linear solution for the endpoint not available")
        use_linear_endpoint = False

    if init_path is not None:
        x_init[1 : len(init_path)] = init_path[1:]

    zshock = np.zeros(len(shocks))
    tshock = np.copy(zshock)
    if shock is not None:
        tshock[shocks.index(shock[0])] = shock[1]

    endpoint = x_lin[-1] if use_linear_endpoint else stst
    out = np.empty((horizon-1,nvars))

    if use_numba in ('p','parallel'):
        stacked_func = lambda x: stacked_func_njit_parallel(x, x0, endpoint, func, horizon, nvars, stst, tshock, zshock, pars)
    elif use_numba:
        stacked_func = lambda x: stacked_func_njit(x, x0, endpoint, func, horizon, nvars, stst, tshock, zshock, pars)
    else:
        stacked_func = lambda x: stacked_func_plain(x, x0, endpoint, func, horizon, nvars, stst, tshock, zshock, pars)

    res = so.root(stacked_func, x_init[1:-1].flatten())

    x[1:-1] = res["x"].reshape((horizon - 1, nvars))

    mess = " ".join(res["message"].replace("\n", " ").split())

    if verbose:
        duration = np.round(time.time() - st, 3)
        print("(find_path_stacked:) Stacking done after %s seconds. " % duration + mess)

    return x, x_lin, not res["success"]



#!/bin/python
# -*- coding: utf-8 -*-

import yaml
import re
import os
import tempfile
import numpy as np
import cloudpickle as cpickle
from copy import deepcopy
from numpy import log, exp, sqrt
from numba import njit
from .steady_state import solve_stst, solve_linear

# initialize model cache
cached_mdicts = ()
cached_models = ()


def eval_strs(vdict):

    if vdict is None:
        return None

    for v in vdict:
        if isinstance(vdict[v], str):
            vdict[v] = eval(vdict[v])

    return vdict


def parse(mfile):
    """parse from yaml file"""

    f = open(mfile)
    mtxt = f.read()
    f.close()

    mtxt = mtxt.replace("^", "**")
    mtxt = re.sub(r"@ ?\n", " ", mtxt)

    # get dict
    model = yaml.safe_load(mtxt)
    # create nice shortcuts
    model["pars"] = model["parameters"]
    model["vars"] = model["variables"]

    return model


def load(model, raise_errors=True, verbose=True):
    """load model from dict or yaml file. Warning: contains filthy code (eg. globals, exec, ...)"""

    global cached_mdicts, cached_models

    if isinstance(model, str):
        model = parse(model)

    if model in cached_mdicts:
        model = cpickle.loads(cached_models[cached_mdicts.index(model)])
        print("(parse:) Loading cached model.")
        return model

    mdict_raw = deepcopy(model)

    defs = model.get("definitions")
    if defs is not None:
        for d in defs:
            exec(d, globals())

    evars = model["variables"]
    dubs = [x for i, x in enumerate(evars) if x in evars[:i]]
    dubmess = (
        ", variables list contains dublicate(s): %s" % ", ".join(dubs) if dubs else ""
    )

    evars = model["variables"][:] = sorted(list(set(evars)), key=str.lower)
    eqns = model["equations"].copy()

    if len(evars) != len(eqns):
        raise Exception(
            "Model has %s variables but %s equations%s."
            % (len(evars), len(eqns), dubmess)
        )
    elif dubs:
        print("(parse:) Warning%s" % dubmess)

    shocks = model.get("shocks") or ()
    par = eval_strs(model["parameters"])

    initvals = model["steady_state"].get("init_guesses")
    stst = eval_strs(model["steady_state"].get("fixed_values"))
    model["stst"] = stst

    # collect number of foward and backward looking variables
    model["no_fwd"] = sum(var + "Prime" in "".join(model["equations"]) for var in evars)
    model["no_bwd"] = sum(var + "Lag" in "".join(model["equations"]) for var in evars)

    # check if each variable is defined in time t (only defining xSS does not give a valid root)
    for v in evars:
        v_in_eqns = [
            v in e.replace(v + "SS", "").replace(v + "Lag", "").replace(v + "Prime", "")
            for e in eqns
        ]
        if not np.any(v_in_eqns):
            raise Exception("Variable `%s` is not defined for time t." % v)

    # start compiling root_container
    for i, eqn in enumerate(eqns):
        if "=" in eqn:
            lhs, rhs = eqn.split("=")
            eqns[i] = "root_container[%s] = " % i + lhs + "- (" + rhs + ")"
        else:
            eqns[i] = "root_container[%s] = " % i + eqn

    eqns_aux = model.get("aux_equations")
    stst_eqns_aux = model.get("steady_state_aux_equations")

    if not shocks:
        shock_str = ""
    elif len(shocks) > 1:
        shock_str = ", ".join(shocks) + " = shocks"
    else:
        shock_str = shocks[0] + " = shocks[0]"

    func_str = """def func_raw(XLag, X, XPrime, XSS, shocks, pars, stst=False):\n %s\n %s\n %s\n %s\n %s\n %s\n root_container=np.empty(%s)\n %s\n %s\n %s\n return root_container""" % (
        ", ".join(v + "Lag" for v in evars) + " = XLag",
        ", ".join(evars) + " = X",
        ", ".join(v + "Prime" for v in evars) + " = XPrime",
        ", ".join(v + "SS" for v in evars) + " = XSS",
        shock_str,
        ", ".join(par.keys()) + " = pars",
        str(len(evars)),
        "if stst:\n  " + "\n  ".join(stst_eqns_aux) if stst_eqns_aux else "",
        "\n ".join(eqns_aux) if eqns_aux else "",
        "\n ".join(eqns),
    )

    # get inital values to test the function
    init = np.ones(len(evars)) * 1.1

    if isinstance(initvals, dict):
        for v in initvals:
            init[evars.index(v)] = initvals[v]

    for v in stst:
        init[evars.index(v)] = stst[v]

    model["init"] = init

    # use a termporary file to get nice debug traces if things go wrong
    tmpf = tempfile.NamedTemporaryFile(mode="w", delete=False)

    tmpf.write(func_str)
    tmpf.close()

    exec(compile(open(tmpf.name).read(), tmpf.name, "exec"), globals())

    # try if function works on initvals. If it does, jit-compile it and remove tempfile
    test = func_raw(
        init, init, init, init, np.zeros(len(shocks)), np.array(list(par.values()))
    )
    if np.isnan(test).any():
        raise Exception("Initial values are NaN.")
    if np.isinf(test).any():
        raise Exception("Initial values are not finite.")

    func = njit(func_raw)
    os.unlink(tmpf.name)

    model["func"] = func
    model["func_str"] = func_str
    model["root_options"] = {}

    if verbose:
        print("(parse:) Parsing done.")

    solve_stst(model, raise_error=raise_errors, verbose=verbose)
    solve_linear(model, raise_error=raise_errors, verbose=verbose)

    cached_mdicts += (mdict_raw,)
    cached_models += (cpickle.dumps(model, protocol=4),)

    return model



#!/bin/python
# -*- coding: utf-8 -*-

from .__init__ import *


def test_bh():

    mod = load(example_bh)

    state = np.zeros(len(mod["variables"]))
    state[:-1] = [0.1, 0.2, 0.0]

    x, _, flag = find_path(mod, state, T=1000, max_horizon=1000, tol=1e-8, verbose=2)

    assert flag == 0
    assert np.allclose(x[9], np.array([0.22287535, 0.25053816, 0.24429734, 0.23821162]))


def test_nk():

    mod = load(example_nk)

    state = mod["stst"].copy()
    state["beta"] *= 1.02

    x, _, flag = find_path(mod, state.values(), verbose=2)

    assert flag == 0
    assert np.allclose(
        x[9],
        np.array(
            [
                1.00608913,
                3.08899444,
                1.00390525,
                1.0,
                0.99659073,
                0.83039318,
                3.08915869,
            ]
        ),
    )


def test_stacked():

    mod = load(example_nk)

    shk = ("e_beta", 0.02)

    x, _, flag = find_path_stacked(mod, shock=shk)

    assert flag == 0
    assert np.allclose(
        x[9],
        np.array(
            [1.00703268, 3.08098172, 1.00377032, 1.0, 0.99306854, 0.82840692, 3.08119]
        ),
    )



#!/bin/python
# -*- coding: utf-8 -*-

import numpy as np
import matplotlib.pyplot as plt
from numba import njit
from interpolation import interp


@njit(cache=True, fastmath=True)
def egm_ces_1asset(bs, par, max_iter=1000, tol=1e-8):

    rb, rl, sigma, beta, prob_g, prob_b, inc_g, inc_b = par
    x = np.vstack((bs, bs))
    cnt = 0
    flag = False

    rPrime = np.ones_like(bs) * rl
    rPrime[bs < 0] = rb
    r = np.empty_like(x)

    while True:

        cnt += 1

        if cnt > max_iter:
            flag = True
            break

        x_old = x.copy()

        fg = interp(x[0], bs, bs)
        fb = interp(x[1], bs, bs)

        cg = (
            beta
            * rPrime
            * (
                prob_g * (inc_g - fg + rPrime * bs) ** -sigma
                + (1 - prob_g) * (inc_b - fb + rPrime * bs) ** -sigma
            )
        ) ** (1 / -sigma)
        cb = (
            beta
            * rPrime
            * (
                prob_b * (inc_g - fg + rPrime * bs) ** -sigma
                + (1 - prob_b) * (inc_b - fb + rPrime * bs) ** -sigma
            )
        ) ** (1 / -sigma)

        x[0] = bs + cg - inc_g
        x[1] = bs + cb - inc_b

        r[:] = rb
        r[0][x[0] >= 0] = rl
        r[1][x[1] >= 0] = rl

        x /= r

        if np.abs(x - x_old).max() < tol:
            break

    c = np.vstack((cg, cb))

    return x, c, flag


@njit(cache=True, fastmath=True)
def find_stst_dist(bins, x, fx, probs, max_iter=1000, tol=1e-8, weights=None):
    """find stationary distribution of 2-state markov chain by iterating on the distribution.
    This is way faster than constructing the transition matrix and calculating the stationary distribution via unity eigenvalues!
    """

    n = len(bins)
    tmat = np.vstack((np.array(probs), 1 - np.array(probs)))

    if weights is None:
        wths = np.ones((2, n)) / n
        wths *= stationary_dist_mc(probs).reshape(-1, 1)
    else:
        wths = weights

    new_bins = np.empty(n + 1)
    # chose new bins such that the original bins are their midpoints
    new_bins[1:-1] = bins[:-1] + np.diff(bins) / 2
    # first and last bin catch everything that is off-grid (they won't carry any weight)
    new_bins[0] = -np.inf
    new_bins[-1] = np.inf

    grid_g = interp(x[0], fx, bins)
    grid_b = interp(x[1], fx, bins)
    bin_inds_g = np.searchsorted(grid_g, new_bins)
    bin_inds_b = np.searchsorted(grid_b, new_bins)

    cnt = 0
    flag = False

    while True:

        cnt += 1

        if cnt > max_iter:
            flag = True
            break

        old_wths = wths.copy()
        wths = tmat @ wths

        cum_wths_g = np.hstack((np.zeros(1), wths[0].cumsum()))
        cum_wths_b = np.hstack((np.zeros(1), wths[1].cumsum()))
        wths[0, :] = np.diff(cum_wths_g[bin_inds_g])
        wths[1, :] = np.diff(cum_wths_b[bin_inds_b])

        if np.abs(wths - old_wths).max() < tol:
            break

    return wths, flag


@njit(cache=True, fastmath=True)
def stationary_dist_mc(*probs):

    probs = np.array(probs)
    mat = np.vstack((probs, 1 - probs))
    v, w = np.linalg.eig(mat)

    return np.ravel(w[:, v == 1] / np.sum(w[:, v == 1]))



from setuptools import setup, find_packages
from os import path

# read the contents of the README file
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/gboehl/econsieve",
    name="econpizza",
    version="0.0.4",
    author="Gregor Boehl",
    author_email="admin@gregorboehl.com",
    description="Solve nonlinear perfect foresight models",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
    ],
    include_package_data=True,
    packages=find_packages(),
    package_data={"pydsge": ["examples/*"]},
    install_requires=[
        "numba",
        "numpy",
        "scipy",
        "grgrlib",
    ],
)
